import re
from performanceplatform.collector.ga.plugins.department import \
    try_get_department

from performanceplatform.collector.ga import \
    create_client, query_documents_for

from performanceplatform.utils.data_pusher import Pusher


def main(credentials, data_set_config, query, options, start_at, end_at):
    client = create_client(credentials)

    assert "filtersets" in options, "`filtersets` must be specified"

    documents = []

    original_filters = query.get("filters", [])

    def insert_ga(filters):
        # Insert ga: on everything but the first.
        def gen():
            for i, f in enumerate(filters):
                if i == 0:
                    yield f
                else:
                    yield "ga:" + f

        return list(gen())

    for filters in options["filtersets"]:

        all_filters = insert_ga(original_filters + filters)
        query["filters"] = [";".join(all_filters)]

        options.setdefault("additionalFields", {}).update(
            {"department": get_department(filters)})

        ds = query_documents_for(
            client, query, options, data_set_config['data-type'],
            start_at, end_at
        )

        documents.extend(ds)

    Pusher(data_set_config, options).push(documents)


def get_department(filters):
    for f in filters:
        try:
            filter_key, filter_value = re.split('=~\^', f)
            return try_get_department(filter_value)
        except:
            raise ValueError("department not found in filters, expected "
                             "filter expression expression containing "
                             "'=~^' e.g. 'Organisation=~^<A1>'.")
    else:
        # If no matches, raise a RuntimeError
        raise RuntimeError("department not found in filters, expected "
                           "filter expression beginning "
                           "'customVarValue9=~^'.")



import os
import logging
import importlib

from collections import OrderedDict

from performanceplatform.collector import arguments
from performanceplatform.collector.logging_setup import (
    set_up_logging, close_down_logging)
from performanceplatform.utils.collector import get_config


def _get_data_group(query):
    return query['data-set']['data-group']


def _get_data_type(query):
    return query['data-set']['data-type']


def _get_data_group_data_type(query):
    return '{}/{}'.format(_get_data_group(query), _get_data_type(query))


def _get_query_params(query):
    """
    >>> _get_query_params({'query': {'a': 1, 'c': 3, 'b': 5}})
    'a=1 b=5 c=3'
    """
    query_params = OrderedDict(sorted(query['query'].items()))
    return ' '.join(['{}={}'.format(k, v) for k, v in query_params.items()])


def _get_path_to_json_file(query):
    return query['path_to_json_file']


def make_extra_json_fields(args):
    """
    From the parsed command-line arguments, generate a dictionary of additional
    fields to be inserted into JSON logs (logstash_formatter module)
    """
    extra_json_fields = {
        'data_group': _get_data_group(args.query),
        'data_type': _get_data_type(args.query),
        'data_group_data_type': _get_data_group_data_type(args.query),
        'query': _get_query_params(args.query),
    }
    if "path_to_json_file" in args.query:
        extra_json_fields['path_to_query'] = _get_path_to_json_file(args.query)
    return extra_json_fields


def logging_for_entrypoint(
        entrypoint, json_fields, logfile_path, logfile_name):
    if logfile_path is None:
        logfile_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), '..', '..', 'log')
    loglevel = getattr(logging, os.environ.get('LOGLEVEL', 'INFO').upper())
    set_up_logging(
        entrypoint, loglevel, logfile_path, logfile_name, json_fields)


def _log_collector_instead_of_running(entrypoint, args):
    logged_args = {
        'start_at': args.start_at,
        'end_at': args.end_at,
        'query': {k: args.query[k] for k in ('data-set', 'query', 'options')}
    }
    logging.info(
        'Collector {} NOT run with the following {}'.format(entrypoint,
                                                            logged_args))


def merge_performanceplatform_config(
        performanceplatform, data_set, token, dry_run=False):
    return {
        'url': '{0}/{1}/{2}'.format(
            performanceplatform['backdrop_url'],
            data_set['data-group'],
            data_set['data-type']
        ),
        'token': token['token'],
        'data-group': data_set['data-group'],
        'data-type': data_set['data-type'],
        'dry_run': dry_run
    }


def _run_collector(entrypoint, args, logfile_path=None, logfile_name=None):
    if args.console_logging:
        logging.basicConfig(level=logging.INFO)
    else:
        logging_for_entrypoint(
            entrypoint,
            make_extra_json_fields(args),
            logfile_path,
            logfile_name
        )

    if os.environ.get('DISABLE_COLLECTORS', 'false') == 'true':
        _log_collector_instead_of_running(entrypoint, args)
    else:
        entrypoint_module = importlib.import_module(entrypoint)
        logging.info('Running collection into {}/{}'.format(
            args.query.get('data-set')['data-group'],
            args.query.get('data-set')['data-type']))
        entrypoint_module.main(
            args.credentials,
            merge_performanceplatform_config(
                args.performanceplatform,
                args.query['data-set'],
                args.token,
                args.dry_run
            ),
            args.query['query'],
            args.query['options'],
            args.start_at,
            args.end_at
        )
        if not args.console_logging:
            close_down_logging()


def main():
    args = arguments.parse_args('Performance Platform Collector')
    if args.collector_slug:
        args.query = get_config(args.collector_slug, args.performanceplatform)

    _run_collector(args.query['entrypoint'], args)


if __name__ == '__main__':
    main()



from performanceplatform.utils.http_with_backoff import HttpWithBackoff
from gapy.client import (
    from_private_key,
    from_secrets_file,
    from_credentials_db,
)
from performanceplatform.utils import statsd


def track_ga_api_usage(kwargs):
    statsd.incr('ga.core.{}.count'.format(kwargs['ids'].replace(':', '')))


def create_client(credentials):
    if "CLIENT_SECRETS" in credentials and "STORAGE_PATH" in credentials:
        return from_secrets_file(
            credentials['CLIENT_SECRETS'],
            storage_path=credentials['STORAGE_PATH'],
            http_client=HttpWithBackoff(),
            ga_hook=track_ga_api_usage,
        )
    elif "ACCOUNT_NAME" in credentials:
        return from_private_key(
            credentials['ACCOUNT_NAME'],
            private_key_path=credentials['PRIVATE_KEY'],
            storage_path=credentials['STORAGE_PATH'],
            http_client=HttpWithBackoff(),
            ga_hook=track_ga_api_usage,
        )
    else:
        return from_credentials_db(
            credentials['CLIENT_SECRETS']['installed'],
            credentials['OAUTH2_CREDENTIALS'],
            http_client=HttpWithBackoff(),
            ga_hook=track_ga_api_usage,
        )



import re
from datetime import datetime
import pytz

from performanceplatform.collector.piwik.base import BaseFetcher
from performanceplatform.utils.data_pusher import Pusher
from performanceplatform.utils.data_parser import DataParser

FREQUENCY_TO_PERIOD_MAPPING = {
    'daily': 'day',
    'weekly': 'week',
    'monthly': 'month',
}


class Fetcher(BaseFetcher):
    def __init__(self, credentials, query, start_at, end_at):
        super(Fetcher, self).__init__(credentials, query)
        self.period = FREQUENCY_TO_PERIOD_MAPPING[
            query.get('frequency', 'weekly')]
        self.date = Fetcher.set_piwik_date(start_at, end_at)
        self.api_method_arguments = query.get('api_method_arguments', None)

    @staticmethod
    def set_piwik_date(start_at, end_at):
        if start_at and end_at:
            date = "{start},{end}".format(
                start=start_at.strftime("%Y-%m-%d"),
                end=end_at.strftime("%Y-%m-%d"))
        else:
            date = "previous1"
        return date

    def _request_params(self):
        params = super(Fetcher, self)._request_params()
        params.update({
            'period': self.period,
            'date': self.date})
        if self.api_method_arguments:
            params.update(self.api_method_arguments)
        return params


class Parser():
    def __init__(self, query, options, data_type):
        self.options = options
        self.data_type = data_type
        self.period = FREQUENCY_TO_PERIOD_MAPPING[
            query.get('frequency', 'weekly')]

    def parse(self, data):
        base_items = []
        special_fields = []
        for date_key, data_points in data.items():
            if type(data_points) == dict:
                data_points = [data_points]
            data = self._parse_item(date_key, data_points)
            base_items += data[0]
            special_fields += data[1]
        return DataParser(
            base_items, self.options, self.data_type
        ).get_data(special_fields)

    def _parse_item(self, date_key, data_points):
        base_items = []
        special_fields = []
        for data_point in data_points:
            start_date = Parser.to_datetime(date_key)
            base_items.append({'start_date': start_date})
            mapped_fields = {
                v: data_point[k] for k, v in self.options['mappings'].items()}
            mapped_fields['timeSpan'] = self.period
            special_fields.append(mapped_fields)
        return base_items, special_fields

    @staticmethod
    def to_datetime(date_key):
        '''
        Extract the first date from 'key' matching YYYY-MM-DD
        or YYYY-MM, and convert to datetime.
        '''
        match = re.search(r'\d{4}-\d{2}(-\d{2})?', date_key)
        formatter = '%Y-%m'
        if len(match.group()) == 10:
            formatter += '-%d'
        return datetime.strptime(
            match.group(), formatter).replace(tzinfo=pytz.UTC)


def main(credentials, data_set_config, query, options, start_at, end_at):
    data_type = data_set_config['data-type']

    data = Fetcher(credentials, query, start_at, end_at).fetch()

    parsed_data = Parser(query, options, data_type).parse(data)

    Pusher(data_set_config, options).push(parsed_data)



from datetime import datetime, time, timedelta, date
from dateutil.relativedelta import relativedelta
import pytz


MONDAY = 0


def to_datetime(a_date):
    if type(a_date) == datetime:
        return a_date.replace(tzinfo=pytz.UTC)
    return datetime.combine(a_date, time(0)).replace(tzinfo=pytz.UTC)


def to_date(a_datetime):
    if a_datetime is None:
        return None
    elif isinstance(a_datetime, datetime):
        return a_datetime.date()
    elif isinstance(a_datetime, date):
        return a_datetime
    else:
        raise ValueError("{0!r} ({1}) isn't a date or datetime"
                         .format(a_datetime, type(a_datetime)))


def to_utc(a_datetime):
    return a_datetime.astimezone(pytz.UTC)


def period_range(start, end, frequency):
    start_date = to_date(start)
    end_date = to_date(end)
    if start_date > end_date:
        raise ValueError("Bad dates: !(start_date={0} <= end_date={1})"
                         .format(start_date, end_date))
    if frequency == 'daily':
        start_date = start_date or a_day_ago()
        end_date = end_date or a_day_ago()
        period = timedelta(days=1)
        return generate_date_range(start_date, end_date, period)
    elif frequency == 'weekly':
        start_date = start_date or a_week_ago()
        start_date = start_of_week(start_date)
        end_date = end_date or a_week_ago()
        period = timedelta(days=7)
        return generate_date_range(start_date, end_date, period)
    elif frequency == 'monthly':
        start_date = start_date or a_month_ago()
        start_date = start_of_month(start_date)
        end_date = end_date or a_month_ago()
        period = relativedelta(months=+1)
        return generate_date_range(start_date, end_date, period)
    else:
        raise ValueError('Bad value of frequency, should be daily, weekly '
                         'or monthly')


def generate_date_range(start_date, end_date, period):
    while start_date <= end_date:
        next_start_date = start_date + period
        yield (to_date(start_date),
               to_date(next_start_date - timedelta(days=1)))
        start_date = next_start_date


def a_week_ago():
    return date.today() - timedelta(days=7)


def a_day_ago():
    return date.today() - timedelta(days=1)


def an_hour_ago():
    return datetime.now() - timedelta(hours=1)


def start_of_month(date):
    return date.replace(day=1)


def a_month_ago():
    return date.today() + relativedelta(months=-1)


def start_of_week(date):
    if date.weekday != MONDAY:
        return date - timedelta(days=date.weekday())
    return date



from performanceplatform.utils import requests_with_backoff
from datetime import datetime
import pytz
from performanceplatform.collector.webtrends.base import(
    BaseCollector, BaseParser)


class Collector(BaseCollector):
    def __init__(self, credentials, query, start_at, end_at):
        self.api_version = credentials.get('api_version')
        self.base_url = credentials['keymetrics_url']
        super(Collector, self).__init__(credentials, query, start_at, end_at)

    @classmethod
    def date_range_for_webtrends(cls, start_at=None, end_at=None):
        """
        Get the start and end formatted for query
        or the last hour if none specified.
        Unlike reports, this does not aggregate periods
        and so it is possible to just query a range and parse out the
        individual hours.
        """
        if start_at and end_at:
            start_date = cls.parse_standard_date_string_to_date(
                start_at)
            end_date = cls.parse_standard_date_string_to_date(
                end_at)
            return [(
                cls.parse_date_for_query(start_date),
                cls.parse_date_for_query(end_date))]
        else:
            return [("current_hour-1", "current_hour-1")]

    def _make_request(self, start_at_for_webtrends, end_at_for_webtrends):
        return requests_with_backoff.get(
            url="{base_url}".format(
                base_url=self.base_url),
            auth=(self.user, self.password),
            params={
                'start_period': start_at_for_webtrends,
                'end_period': end_at_for_webtrends,
                'format': self.query_format,
                "userealtime": True
            }
        )

    def build_parser(self, data_set_config, options):
        if self.api_version == 'v3':
            return V3Parser(options, data_set_config['data-type'])
        else:
            return V2Parser(options, data_set_config['data-type'])


class V2Parser(BaseParser):
    def parse_item(self, item):
        initial_data = []
        special_fields = []
        for key, value in item.items():
            if value['SubRows']:
                row_items = value['SubRows'].items()
                for date, row in row_items:
                    start_date = self.__class__.to_datetime(date)
                    if row:
                        initial_data.append({
                            'start_date': start_date,
                        })
                        special_fields.append(row['measures'])
        return initial_data, special_fields

    @classmethod
    def to_datetime(cls, webtrends_formatted_date):
        return datetime.strptime(
            webtrends_formatted_date, "%m/%d/%Y %H").replace(tzinfo=pytz.UTC)


class V3Parser(BaseParser):
    def parse_item(self, items):
        initial_data = []
        special_fields = []
        for value in items:
            start_date = self.__class__.to_datetime(value)
            if value['SubRows']:
                for _, row in value['SubRows'].items():
                    if row:
                        initial_data.append({
                            'start_date': start_date,
                        })
                        special_fields.append(row['measures'])
        return initial_data, special_fields

    @classmethod
    def to_datetime(cls, webtrends_formatted_data_point):
        start_at = webtrends_formatted_data_point['start_date']
        date_parts = start_at.split('-')
        if len(date_parts) == 3:
            return datetime.strptime(
                start_at, "%Y-%m-%d").replace(tzinfo=pytz.UTC)
        else:
            return datetime.strptime(
                start_at, "%Y-%m").replace(tzinfo=pytz.UTC)


def main(credentials, data_set_config, query, options, start_at, end_at):
    collector = Collector(credentials, query, start_at, end_at)
    collector.collect_parse_and_push(data_set_config, options)



from performanceplatform.utils import requests_with_backoff


class BaseFetcher(object):
    def __init__(self, credentials, query):
        self.token_auth = credentials['token_auth']
        self.url = credentials['url']
        self.api_method = query['api_method']
        self.site_id = query['site_id']
        self.query_format = 'json'

    def _request_params(self):
        return {
            'method': self.api_method,
            'idSite': self.site_id,
            'format': self.query_format,
            'token_auth': self.token_auth
        }

    def _make_request(self):
        return requests_with_backoff.get(
            url="{url}?module=API".format(url=self.url),
            params=self._request_params()
        )

    def fetch(self):
        response = self._make_request()
        response.raise_for_status()
        return response.json()



from performanceplatform.client import DataSet
import logging


class Pusher(object):
    def __init__(self, target_data_set_config, options):
        self.data_set_client = DataSet.from_config(target_data_set_config)
        self.chunk_size = options.get('chunk-size', 100)
        self.empty_data_set = options.get('empty-data-set', False)

    def push(self, data):
        if data:
            if self.empty_data_set:
                self.data_set_client.empty_data_set()
            self.data_set_client.post(
                data, chunk_size=self.chunk_size)
        else:
            logging.info("Doing nothing - no data to push")



from datetime import datetime
import pytz
from performanceplatform.utils import requests_with_backoff
from performanceplatform.collector.logging_setup import (
    extra_fields_from_exception )
import time
import logging


class Pingdom(object):
    def __init__(self, config):
        self.user = config['user']
        self.password = config['password']
        self.app_key = config['app_key']
        self.API_LOCATION = "https://api.pingdom.com/api/2.0/"

    def _make_request(self, path, url_params=None):
        if url_params is None:
            url_params = {}
        response = requests_with_backoff.get(
            url=self.API_LOCATION + path,
            auth=(self.user, self.password),
            headers={
                "App-Key": self.app_key
            },
            params=url_params
        )
        response.raise_for_status()

        return response.json()

    def _build_response(self, data):
        hours = data['summary']['hours']
        new_hours = []
        for hour in hours:
            hour.update({'starttime': datetime.fromtimestamp(
                hour['starttime'],
                tz=pytz.UTC
            )})
            new_hours.append(hour)
        return new_hours

    def stats(self, check_name, start, end):
        app_code = self.check_id(check_name)
        params = {
            "includeuptime": "true",
            "from": time.mktime(start.timetuple()),
            "to": time.mktime(end.timetuple()),
            "resolution": "hour"
        }
        path = "summary.performance/" + str(app_code)

        try:
            return self._build_response(self._make_request(
                path=path,
                url_params=params
            ))
        except requests_with_backoff.exceptions.HTTPError as e:
            logging.error("Request to pingdom failed: %s" % str(e),
                          extra=extra_fields_from_exception(e))

    def check_id(self, name):
        checks = self._make_request(path="checks")

        check_to_find = [check for check in checks["checks"]
                         if check["name"] == name]

        return check_to_find[0]["id"]
